/*
	Avr.C
	
	Top class of the AVR micro controllers 
	Uros Platise (c) 1999
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "Avr.h"
#include "Error.h"

#define TARGET_MISSING 0xff
#define DEVICE_LOCKED  0x1

/* ATMEL AVR codes */

TAvr::TPart TAvr::parts [] = {
  {"ATmega103-old", 0x01, 0x01, 0x20000, 0x1000,  56000, 22000, 9000, 4000}, 
  {"ATmega103",     0x97, 0x01, 0x20000, 0x1000,  56000, 22000, 9000, 4000}, 
  {"ATmega603",     0x96, 0x01, 0x10000, 0x1000,  56000, 22000, 9000, 4000}, 

  {"AT90S1200", 0x90, 0x01, 1024, 64, 18000, 8000, 9000, 4000}, /* delay? */
  
  /* QuickFlash(TM) - OTP? */
  {"ATtiny10",  0x90, 0x03, 1024,  0, 18000, 8000, 9000, 4000}, /* delay? */
  
  /* 12V serial programming */
  {"ATtiny11",  0x90, 0x04, 1024,   0, 18000, 8000, 9000, 4000}, 
  {"ATtiny12",  0x90, 0x05, 1024,  64, 18000, 8000, 9000, 4000},

  {"AT90S2313", 0x91, 0x01, 2048, 128, 18000, 8000, 9000, 4000},
  {"AT90S2343", 0x91, 0x03, 2048, 128, 18000, 8000, 9000, 4000}, /* 97 ? */
  {"AT90S2323", 0x91, 0x04, 2048, 128, 18000, 8000, 9000, 4000}, /* 97 ? */  
  {"AT90S2333", 0x91, 0x05, 2048, 128, 18000, 8000, 9000, 4000},
  {"ATtiny22",  0x91, 0x06, 2048, 128, 18000, 8000, 9000, 4000},

  {"AT90S4414", 0x92, 0x01, 4096, 256, 18000, 8000, 9000, 4000},
  {"AT90S4433", 0x92, 0x03, 4096, 256, 18000, 8000, 9000, 4000},
  {"AT90S4434", 0x92, 0x02, 4096, 256, 18000, 8000, 9000, 4000},
  
  {"AT90S8515", 0x93, 0x01, 8192, 512, 18000, 8000, 9000, 4000},
  
  /* parallel programming only; here just for the evidence */
  {"AT90C8534", 0x93, 0x04, 8192, 512, 18000, 8000, 9000, 4000}, 
    
  {"AT90S8535", 0x93, 0x03, 8192, 512, 18000, 8000, 9000, 4000},

  {"",          TARGET_MISSING},
  {"locked",    DEVICE_LOCKED},
  {"",          0x0}
};

const char* TAvr::segment_names[] = {"flash", "eeprom", "fuse", NULL};


/* Private Functions
*/

bool TAvr::IsWritePageModeProvided(){
  if (device_locked){return false;}
  assert(part!=NULL);
  return strncmp(part->name, "ATmega", 6)==0;
}

/* Protected Functions
*/

void TAvr::Identify(){
  char* vendor = "Device";
  
  Info(3, "Vendor Code: %x\nPart Family: %x\nPart Number: %x\n",
    vendor_code, part_family, part_number);  
  
  /* Identify AVR Part according to the vendor_code ... */
  if (vendor_code==0x1e){vendor = "Atmel AVR";}
  
  if (vendor_code==0 && part_family==DEVICE_LOCKED && part_number==0x02){
    device_locked=true;
    Info(0, "Cannot identify device because is it locked.\n");
    return;
  } else{device_locked=false;}
  if (part_family==TARGET_MISSING){
    Info(0, "An error has occuried durring the AVR initilization.\n"
	    " * Target status:\n"
	    "   Vendor Code = %x, Part Family = %x, Part Number = %x\n\n",
	    vendor_code, part_family, part_number);
    throw 
      Error_Device("Probably the wiring is incorrect or target"
        " might be `damaged'.");
  }
  int i,n;
  for(i=0; parts[i].part_family != 0x0; i++){
    if (part_family == parts[i].part_family){
      for (n=i; parts[n].part_number==parts[i].part_number; n++){
        if (part_number == parts[n].part_number){i=n; break;}
      }
      if (i==n){Info(1, "%s %s is found.\n", vendor, parts[i].name);}
      else{Info(1, "%s similar to the %s is found.\n", vendor, parts[i].name);}
      part = &parts[i];
      break;
    }
  }
  if (parts[i].part_family == 0x0) {
    throw Error_Device ("Probably the AVR MCU is not in the RESET state.\n"
			"Check it out and run me again.");}
			
  page_mode_enabled = IsWritePageModeProvided();
  Info(3, "Page Mode Enabled: %s\n", page_mode_enabled?"Yes":"No");
  
  SetWriteTimings();
}

#define CALC_FLASH_T_wd(voltage) (long)(part->t_wd_flash_32 + (voltage - 3.2)*k_flash)
#define CALC_EEPROM_T_wd(voltage) (long)(part->t_wd_eeprom_32 + (voltage - 3.2)*k_eeprom)

void TAvr::SetWriteTimings(){
  const char* val;
  long k_flash = 
    (long)((part->t_wd_flash_50 - part->t_wd_flash_32)/AVR_VOLTAGE_SPRANGE);
    
  long k_eeprom = 
    (long)((part->t_wd_eeprom_50 - part->t_wd_eeprom_32)/AVR_VOLTAGE_SPRANGE);
    
  /* defaults */
  t_wd_flash = CALC_FLASH_T_wd(AVR_DEFAULT_VOLTAGE);
  t_wd_eeprom = CALC_EEPROM_T_wd(AVR_DEFAULT_VOLTAGE);  
    
  /* set FLASH write delay */  
  if ((val=GetCmdParam("-dt_wd_flash"))){
    t_wd_flash = atol(val);
    Info(0, "t_wd_flash = %ld\n", t_wd_flash);
    if (t_wd_flash < CALC_FLASH_T_wd(AVR_MAX_VOLTAGE)){
      Info(0, " * According to the Atmel specs the t_wd_flash\n"
              "   should be in range from %ld to %ld us\n", 
	      CALC_FLASH_T_wd(AVR_MIN_VOLTAGE),
	      CALC_FLASH_T_wd(AVR_MAX_VOLTAGE));
	      
      throw Error_Device("-dt_wd_flash: Value out of range.");
    }
  }
  
  /* set EEPROM write delay */  
  if ((val=GetCmdParam("-dt_wd_eeprom"))){
    t_wd_eeprom = atol(val);
    if (t_wd_eeprom < CALC_EEPROM_T_wd(AVR_MAX_VOLTAGE)){
      Info(0, " * According to the Atmel specs the t_wd_eeprom\n"
              "   should be in range from %ld to %ld us\n", 
	      CALC_EEPROM_T_wd(AVR_MIN_VOLTAGE),
	      CALC_EEPROM_T_wd(AVR_MAX_VOLTAGE));
	      
      throw Error_Device("-dt_wd_eeprom: Value out of range.");
    }
  }

  /* Set Timings according to the Power Supply Voltage */
  if ((val=GetCmdParam("-dvoltage"))){
    double voltage = atof(val);
    if (voltage < AVR_MIN_VOLTAGE || voltage > AVR_MAX_VOLTAGE){
      Info(0, " * Atmel AVR MCUs operate in range from %.1f to %.1f V\n",
           AVR_MIN_VOLTAGE, AVR_MAX_VOLTAGE);
	   
      throw Error_Device("-dvoltage: Value out of range.");
    }
    
    t_wd_flash = CALC_FLASH_T_wd(voltage);
    t_wd_eeprom = CALC_EEPROM_T_wd(voltage);
  }
  
  Info(3, "FLASH Write Delay (t_wd_flash): %ld us\n"
          "EEPROM Write Delay (t_wd_eeprom): %ld us\n",
	  t_wd_flash, t_wd_eeprom);
}

const char* TAvr::GetPartName(){
  return part->name;
}

void TAvr::CheckMemoryRange(TAddr addr){
  if (device_locked){
    Info(0, "Device is locked.\n");
    throw Error_MemoryRange();
  }
  if (segment==SEG_FLASH && addr >= part->flash_size ||
      segment==SEG_EEPROM && addr >= part->eeprom_size ||
      segment==SEG_FUSE && addr != 0){
    throw Error_MemoryRange();
  }
}

long TAvr::Get_t_wd_flash() const {
  return t_wd_flash;
}

long TAvr::Get_t_wd_eeprom() const {
  return t_wd_eeprom;
}

long TAvr::Get_t_rst() const{
  return 3*t_wd_flash;	/* right factor is 2, but just in case */
}


/* Device Interface Functions
*/

bool TAvr::SetSegment(const char* segment_name){
  for (int i=0; segment_names[i]!=NULL; i++){
    if (strcmp(segment_names[i], segment_name)==0){
      segment=(TSegment)i; 
      return true;
    }
  }
  return false;
}

const char* TAvr::TellActiveSegment(){
  return segment_names[segment];
}

const char* TAvr::ListSegment(unsigned index){
  if (index>3){return NULL;}
  return segment_names[index];
}

const char* TAvr::ReadByteDescription(TAddr addr){
  static const char* no_desc = "No description available.";
  CheckMemoryRange(addr);
  return no_desc;
}

/* Constructor/Destructor
*/

TAvr::TAvr():
  part(NULL), 
  page_mode_enabled(false), page_addr_fetched(false),
  segment(SEG_FLASH){
}

TAvr::~TAvr(){
}
