/*
	AvrDummy.C

	Dummy device driver for the AVR parallel access
	Uros Platise (c) 1999
*/

#include <sys/time.h>
#include "AvrDummy.h"

/* Private Functions
*/

void TAvrDummy::EnableAvr(){
  unsigned char prg  [4] = { 0xAC, 0x53, 0, 0 };
  int try_number = 32;
  bool no_retry = GetCmdParam("-dno_retry", false);

  /* Enable AVR programming mode */
  do{
    prg[0]=0xAC; prg[1]=0x53; prg[2]=prg[3]=0;
    Send(prg, 4);
    if (no_retry) break;
    if (prg[2] == 0x53) break;
    SCKtoRetryEnable();
  } while (try_number--);

  if (try_number>=0){
    Info(2,"AVR Direct Parallel Access succeeded after %d retries.\n",
      32-try_number);
  }

  /* Get AVR Info */
  vendor_code = GetPartInfo(0);
  part_family = GetPartInfo(1);
  part_number = GetPartInfo(2);
  Identify();
}

TByte TAvrDummy::GetPartInfo(TAddr addr){
  TByte info [4] = { 0x30, 0, addr, 0 };
  Send(info, 4);
  return info[3];
}

void TAvrDummy::WriteProgramMemoryPage(){
  TByte prg_page [4] = { 0x4C,
			(TByte)((page_addr >> 9) & 0xff),
			(TByte)((page_addr >> 1) & 0xff),
			0 };
  Info(4, "Programming page address: %d (%.2x, %.2x, %.2x, %.2x)\n",
    page_addr, prg_page[0], prg_page[1], prg_page[2], prg_page[3]);
  Send(prg_page, 4);

  /* Wait */
  /* tWD_PROG = 9ms at Vcc=3.2V and */
  /* approx. 12 ms at 2.6 V */
  usleep (Get_t_wd_flash());
  page_addr_fetched=false;
}


/* Device Interface Functions
*/

TByte TAvrDummy::ReadByte(TAddr addr){
  struct timeval t0, t2;
  TByte readback = 0;
  gettimeofday(&t0, NULL);

  CheckMemoryRange(addr);
  if (segment==SEG_FLASH){
    TByte hl = (addr&1) ? high_byte : low_byte;
    addr>>=1;
    TByte flash [4] = { 0x20+hl,
			(TByte)((addr >> 8) & 0xff),
			(TByte)(addr & 0xff),
			0 };
    Send(flash, 4);
    readback = flash[3];
  }
  else if (segment==SEG_EEPROM){
    TByte eeprom [4] = { 0xA0,
			 (TByte)((addr>>8)&0xff),
			 (TByte)(addr&0xff),
			 0 };
    Send(eeprom, 4);
    readback = eeprom[3];
  }

  gettimeofday(&t2, NULL);

  /* Read Statistics */
  timersub(&t2, &t0, &t2);
  float rate = 1 / ((float)(t2.tv_sec + t2.tv_usec*1e-6));
  if (rate > max_read_rate){max_read_rate = rate;}
  if (rate < min_read_rate){min_read_rate = rate;}

  return readback;
}

void TAvrDummy::WriteByte(TAddr addr, TByte byte, bool flush_buffer){
  struct timeval t0, t1, t2;
  TByte rbyte=0;
  bool device_not_erased=false;

  /* Poll data if use_data_polling is enabled and if page mode
     is enabled, flash is not selected */
  bool poll_data = !(page_mode_enabled & (segment==SEG_FLASH)) &
                   use_data_polling;

  CheckMemoryRange(addr);

  /* For speed, don't program a byte that is already there
     (such as 0xFF after chip erase).  */
  if (poll_data){
    rbyte=ReadByte(addr);
    if (rbyte == byte){return;}
    if (rbyte != 0xff){device_not_erased=true;}
  }

  t2.tv_sec = 0;

  if (segment==SEG_FLASH){

    /* PAGE MODE PROGRAMMING:
       If page mode is enabled cache page address.
       When current address is out of the page address
       flush page buffer and continue programming.
    */
    if (page_mode_enabled){
      Info(4, "Loading data to address: %d (page_addr_fetched=%s)\n",
        addr, page_addr_fetched?"Yes":"No");

      if (page_addr_fetched && page_addr != (addr & 0x1ff00)){
        WriteProgramMemoryPage();
      }
      if (page_addr_fetched==false){
        page_addr=addr & 0x1ff00;
	page_addr_fetched=true;
      }
      if (flush_buffer){WriteProgramMemoryPage();}
    }

    TByte hl = (addr&1) ? high_byte : low_byte;
    TByte flash [4] = { 0x40+hl,
			(TByte)((addr >> 9) & 0xff),
			(TByte)((addr >> 1) & 0xff),
			byte };
    Send(flash, 4);

    /* We do not need to wait for each byte in page mode programming */
    if (page_mode_enabled){return;}
    t2.tv_usec = Get_t_wd_flash();
  }
  else if (segment==SEG_EEPROM){
    TByte eeprom [4] = { 0xC0,
			 (TByte)((addr>>8)&0xff),
			 (TByte)(addr&0xff),
			 byte };
    Send(eeprom, 4);
    t2.tv_usec = Get_t_wd_eeprom();
  }

  gettimeofday(&t0, NULL);
  gettimeofday(&t1, NULL);
  timeradd(&t1, &t2, &t1);  	/* t1 += tWD_PROG */
  do{
    /* Data Polling: if the programmed value reads correctly, and
       is not equal to any of the possible P1, P2 read back values,
       it is done; else wait until tWD_PROG time has elapsed.
       The busy loop here is to avoid rounding up the programming
       wait time to 10ms timer ticks (for Linux/x86).  Programming
       is not really "hard real time" but 10ms instead of ~4ms for
       every byte makes it slow.  gettimeofday() reads the 8254
       timer registers (or Pentium cycle counter if available),
       so it has much better (microsecond) resolution.
    */
    if (poll_data){
      if ((byte == (rbyte = ReadByte(addr))) &&
          (byte != 0) && (byte != 0x7F) && (byte != 0x80) && (byte != 0xFF)){
	break;
      }
    }
    gettimeofday(&t2, NULL);
  } while (timercmp(&t2, &t1, <));

  /* Write Statistics */
  timersub(&t2, &t0, &t2);
  float rate = 1 / ((float)(t2.tv_sec + t2.tv_usec*1e-6));
  if (rate > max_write_rate){max_write_rate = rate;}
  if (rate < min_write_rate){min_write_rate = rate;}

  if (poll_data && byte != rbyte){
    if (device_not_erased){
      Info(0, "Warning: It seems that device is not erased.\n"
              "         Erase it with the --erase option.\n");
    }
    Info(0, "Error: Data polling readback status: write=$%.2x  read=$%.2x\n",
      byte, rbyte);
    throw Error_Device("If device was erased disable polling with the "
      "-dno-poll option.");
  }
}

void TAvrDummy::FlushWriteBuffer(){
  if (page_addr_fetched){
    WriteProgramMemoryPage();
  }
}

void TAvrDummy::ChipErase(){
  TByte chip_erase [4] = { 0xAC, 0x80, 0x00, 0x00 };
  Send (chip_erase, 4);
  Info(1, "Erasing device ...\n");
  usleep(Get_t_rst());
  Reset();
  Info(1, "Reinitializing device\n");
  EnableAvr();
}

const char* TAvrDummy::GetMaxWriteRate(){
  static char max_wrb[16];
  if (max_write_rate==0){return MinMaxNone();}
  sprintf(max_wrb,"%.0f", max_write_rate);
  return max_wrb;
}

const char* TAvrDummy::GetMinWriteRate(){
  static char min_wrb[16];
  if (min_write_rate==1e6){return MinMaxNone();}
  sprintf(min_wrb,"%.0f", min_write_rate);
  return min_wrb;
}

const char* TAvrDummy::GetMaxReadRate(){
  static char max_rdb[16];
  if (max_read_rate==0){return MinMaxNone();}
  sprintf(max_rdb,"%.0f", max_read_rate);
  return max_rdb;
}

const char* TAvrDummy::GetMinReadRate(){
  static char min_rdb[16];
  if (min_read_rate==1e6){return MinMaxNone();}
  sprintf(min_rdb,"%.0f", min_read_rate);
  return min_rdb;
}

void TAvrDummy::ResetMinMax(){
  max_write_rate = max_read_rate = 0;
  min_write_rate = min_read_rate = 1e6;
}

/* Constructor
*/

TAvrDummy::TAvrDummy():
  use_data_polling(true){

  ResetMinMax();

  /* Device Command line options ... */
  if (GetCmdParam("-dno-poll", false)){use_data_polling=false;}

  /* Not that important stuff */
  if (GetCmdParam("-favr_endianbug", false)){
    high_byte = 0x00;
    low_byte = 0x08;
  } else{
    high_byte = 0x08;
    low_byte = 0x00;
  }

  EnableAvr();
}
/* eof */
