/*
	DAPA.C

	Direct AVR Parallel Access (c) 1999

	Originally written by Sergey Larin.
	Corrected by
	  Denis Chertykov,
	  Uros Platise and
	  Marek Michalkiewicz

	CASC mode by Serge Vakulenko
	(see http://www.vak.ru/software/casc.html)
*/

//#define DEBUG

#include <stdio.h>
#include <stdlib.h>

#ifndef NO_DAPA
/* sys/io.h is the correct file.  If your system doesn't have that
   file, something is broken.
*/
#ifdef __BROKEN_LINUX
#include <asm/io.h>
#else
#include <sys/io.h>
#endif
#include <linux/lp.h>
#endif

#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/time.h>
#include "Global.h"
#include "Error.h"
#include "DAPA.h"
#ifndef WIN32
#include <sys/ioctl.h>
#include <termios.h>
#endif

/* Parallel Port Base Address
*/
#define BASE        parport_base

/* Alex's Direct Avr Parallel Access
*/
#define DAPA_SCK    LP_PSTROBE	/* base + 2 */
#define DAPA_RESET  LP_PINITP	/* base + 2 */
#define DAPA_DIN    LP_PBUSY	/* base + 1 */
#define DAPA_DOUT   0x1		/* base */

/* STK200 Direct Parallel Access
*/
#define STK2_ENA1   0x04    	/* D2 (base) - enable for RESET, MISO */
#define STK2_ENA2   0x08    	/* D3 (base) - enable for SCK, MOSI */
#define STK2_SCK    0x10    	/* D4 (base) */
#define STK2_DOUT   0x20    	/* D5 (base) */
#define STK2_RESET  0x80    	/* D7 (base) */
#define STK2_DIN    0x40    	/* ACK (base + 1) */

/* Altera Byte Blaster Port Configuration
*/
#define ABB_EN      LP_PAUTOLF	/* low active */
#define ABB_LPAD    0x80	/* D7: loop back throught enable auto-detect */
#define ABB_SCK	    0x01	/* D0: TCK */
#define ABB_RESET   0x02	/* D1: TMS */
#define ABB_DOUT    0x40	/* D6: TDI */
#define ABB_DIN	    LP_PBUSY	/* BUSY: TDO */

#ifndef NO_DAPA
static volatile int yes_alarm;
static void alarm( int k ){
  yes_alarm = 0;
}

void TDAPA::DelayFunc(){
  volatile int i;
  for (i=0;i<100;i++);
}

void TDAPA::BigDelay(){
  static volatile int n;
  n = n_per_ms/25000 + 1;
  while (n){DelayFunc();n--;}
}

long TDAPA::CalcDelay(){ // pause x ms
  long n=0;
  struct itimerval it;
  it.it_interval.tv_sec = 0;
  it.it_value.tv_sec = 0;
  it.it_interval.tv_usec = 0;
  it.it_value.tv_usec = 100000;
  signal(SIGALRM, alarm);
  setitimer(ITIMER_REAL, &it, NULL);
  yes_alarm = 1;
  while (yes_alarm){
    DelayFunc(); n++;
  }
  return n;
}
#endif /* NO_DAPA */

void TDAPA::Reset(){
  switch(pa_type){
#ifndef NO_DAPA
  case PAT_DAPA:
    outb(DAPA_SCK, BASE+2); usleep(1000);      /* sck = 0, reset = 0 */
    outb(DAPA_SCK | DAPA_RESET, BASE+2);       /* sck = 0, reset = 1 */
    usleep(1000);
    outb(DAPA_SCK, BASE+2); usleep(1000);      /* sck = 0, reset = 0 */
    break;

  case PAT_STK200:
    outb(STK2_ENA1 | STK2_ENA2, BASE);    	/* leave MCU */
    usleep(30000);
    outb(STK2_ENA2, BASE); usleep(1000);	/* sck = Z, reset = 0 */
    outb(0, BASE); usleep(1000);  		/* sck = 0, reset = 0 */
    outb(STK2_RESET, BASE); usleep(1000); 	/* sck = 0, reset = 1 */
    outb(0, BASE); usleep(1000);         	/* sck = 0, reset = 0 */
    break;

  case PAT_ABB:
    outb(0,BASE+2); usleep(30000);		/* leave MCU */
    outb(0,BASE); 				/* reset=sck=0 */
    outb(ABB_EN,BASE+2); usleep(30000);		/* enable 244 driver */
    outb(ABB_RESET,BASE); usleep(30000);	/* reset=1 */
    outb(0,BASE); usleep(30000);		/* reset=0*/
//    fprintf(stderr, "quiting ... test\n"); exit(1);
    break;
#endif
  case PAT_CASC:
#ifdef WIN32
    EscapeCommFunction (serline, CLRDTR);	/* reset=1 */
    usleep (1000);
    EscapeCommFunction (serline, SETDTR);	/* reset=0 */
#else
    int val = TIOCM_DTR;
    ioctl(serline, TIOCMBIC, &val);		/* reset=1 */
    ioctl(serline, TIOCMBIS, &val);		/* reset=0 */
#endif
    break;
  }
  usleep(20000);
}

void TDAPA::SCKtoRetryEnable(){
  switch(pa_type){
#ifndef NO_DAPA
  case PAT_DAPA:
    outb(0, BASE+2); BigDelay();
    outb(DAPA_SCK, BASE+2); BigDelay();
    break;

  case PAT_STK200:
    outb(STK2_SCK, BASE); BigDelay();  /* sck = 1 */
    outb(0, BASE); BigDelay();  /* sck = 0 */
    break;

  case PAT_ABB:
    outb(ABB_SCK, BASE); BigDelay();
    outb(0, BASE); BigDelay();
    break;
#endif
  case PAT_CASC:
#ifdef WIN32
    EscapeCommFunction (serline, SETRTS);	/* sck=0 */
    EscapeCommFunction (serline, CLRRTS);	/* sck=1 */
#else
    int val = TIOCM_RTS;
    ioctl(serline, TIOCMBIS, &val);		/* sck=0 */
    ioctl(serline, TIOCMBIC, &val);		/* sck=1 */
#endif
    break;
  }
}

unsigned char TDAPA::SendRecv(unsigned char b){
  int i, mask, received=0;

  for (i=0, mask = 0x80; i<8; i++, mask >>= 1){
     switch(pa_type){
#ifndef NO_DAPA
     case PAT_DAPA:
       outb(b & mask ? DAPA_DOUT : 0, BASE); BigDelay();
       outb(0, BASE+2); BigDelay ();
       if (!(inb (BASE+1) & DAPA_DIN)){received |= mask;}
       outb(DAPA_SCK, BASE+2); BigDelay();
       break;

     case PAT_STK200:
       outb((b & mask) ? STK2_DOUT : 0, BASE);  /* send data bit, sck = 0 */
       BigDelay();
       if (inb(BASE + 1) & STK2_DIN){received |= mask;} /* read data bit */
       						/* send data bit, sck = 1 */
       outb(STK2_SCK | ((b & mask) ? STK2_DOUT : 0), BASE);
       BigDelay();
       outb((b & mask) ? STK2_DOUT : 0, BASE);	/* send data bit, sck = 0 */
       BigDelay();
       break;

     case PAT_ABB:
       outb((b & mask) ? ABB_DOUT : 0, BASE);  /* send data bit, sck = 0 */
       BigDelay();
       if (!(inb(BASE + 1) & ABB_DIN)){received |= mask;} /* read data bit */
       						/* send data bit, sck = 1 */
       outb(ABB_SCK | ((b & mask) ? ABB_DOUT : 0), BASE);
       BigDelay();
       outb((b & mask) ? ABB_DOUT : 0, BASE);	/* send data bit, sck = 0 */
       BigDelay();
       break;
#endif
     case PAT_CASC:
#ifdef WIN32
       /* send data bit */
       EscapeCommFunction (serline, (b & mask) ? CLRBREAK : SETBREAK);

       /* send clock */
       EscapeCommFunction (serline, SETRTS);	/* sck=0 */
       EscapeCommFunction (serline, CLRRTS);	/* sck=1 */

       /* read data bit */
       DWORD val;
       GetCommModemStatus (serline, &val);
       if (!(val & MS_CTS_ON)){received |= mask;}
#else
       /* send data bit */
       ioctl (serline, (b & mask) ? TIOCCBRK : TIOCSBRK, 0);
       /* send clock */
       int val = TIOCM_RTS;
       ioctl(serline, TIOCMBIS, &val);		/* sck = 0 */
       ioctl(serline, TIOCMBIC, &val);		/* sck = 1 */
       /* read data bit */
       ioctl (serline, TIOCMGET, &val);
       if (!(val & TIOCM_CTS)){received |= mask;}
#endif
       break;
     }

  }
  return received;
}

int TDAPA::Send (unsigned char* queue, int queueSize, int rec_queueSize=-1){
  unsigned char *p = queue, ch;
  int i = queueSize;

  if (rec_queueSize==-1){rec_queueSize = queueSize;}
#ifdef DEBUG
  printf ("send(recv): ");
#endif
  while (i--){
#ifdef DEBUG
    printf ("%02X(", (unsigned int)*p);
#endif
    ch = SendRecv(*p);
#ifdef DEBUG
    printf ("%02X) ", (unsigned int)ch);
#endif
    *p++ = ch;
  }
#ifdef DEBUG
  printf ("\n");
#endif
  return queueSize;
}

TDAPA::TDAPA():
  parport_base(0x378){

  const char* val;
  /* Parse Command Line Switches */
  if ((val=GetCmdParam("-dlpt"))!=NULL){
    if (!strcmp(val, "1")){parport_base = 0x278;}
    else if (!strcmp(val, "2")){parport_base = 0x378;}
    else if (!strcmp(val, "3")){parport_base = 0x3bc;}
    else{
      parport_base = strtol(val, NULL, 0);
      if (parport_base!=0x278 && parport_base!=0x378 && parport_base!=0x3bc){
        throw Error_Device("Bad device address.");
      }
    }
  }

  /* Enable Parallel Port */
  if (GetCmdParam("-dcasc", false)){pa_type = PAT_CASC;}
#ifndef NO_DAPA
  else if (GetCmdParam("-dapa", false)){pa_type = PAT_DAPA;}
  else if (GetCmdParam("-dstk200", false)){pa_type = PAT_STK200;}
  else if (GetCmdParam("-dabb", false)){pa_type = PAT_ABB;}
#endif
  else{
    throw Error_Device("Direct Parallel Access not defined.");
  }

#ifndef NO_DAPA
  n_per_ms = CalcDelay();
  Info(3, "Number of delay loops per 100 millisecond: %ld\n", n_per_ms);
#endif
  if (pa_type == PAT_CASC){
#ifdef WIN32
    const char* dev_name = "COM2";
#else
    const char* dev_name = "/dev/avr";
#endif
    if ((val=GetCmdParam("-dserial"))){dev_name = val;}
#ifdef WIN32
    serline = CreateFile (dev_name, GENERIC_READ | GENERIC_WRITE,
      0, 0, OPEN_EXISTING, 0, 0);
    if (serline == INVALID_HANDLE_VALUE) {
      printf("Cannot open %s\n", dev_name);
      throw Error_C();
    }
#else
    /* Open port and set serial attributes */
    if ((serline = open(dev_name, O_RDWR | O_NDELAY)) < 0 ){throw Error_C();}
    struct termios pmode;
    tcgetattr(serline, &pmode);
    cfmakeraw (&pmode);
    pmode.c_iflag |= IGNBRK;
    tcsetattr(serline, TCSANOW, &pmode);
#endif
  }
#ifndef NO_DAPA
  else if (::ioperm(BASE, 8, 1) != 0)
    {perror("ioperm()");throw Error_Device("DPA","Cannot open port.");}
#endif
  Reset();
}

TDAPA::~TDAPA(){
  switch(pa_type){
#ifndef NO_DAPA
  case PAT_DAPA:   outb(DAPA_SCK | DAPA_RESET, BASE+2); break;
  case PAT_STK200: outb(STK2_ENA1 | STK2_ENA2, BASE);   break;
  case PAT_ABB:    outb(0, BASE+2); outb(ABB_RESET, BASE); break;
#endif
  case PAT_CASC:
#ifdef WIN32
    CloseHandle(serline);
#else
    close(serline);
#endif
    break;
  }
}
/* eof */
