/*
	Serial.C
	
	Serial Interface
	Uros Platise, (c) 1997-1999
*/

#include <sys/time.h>
#ifndef WIN32
#include <sys/ioctl.h>
#include <termios.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include "Global.h"
#include "Serial.h"

int TSerial::Tx(unsigned char* queue, int queue_size){
#ifdef WIN32
  DWORD count;
  
  WriteFile (serline, queue, queue_size, &count, 0);
  return count;
#else
  return write(serline, queue, queue_size);
#endif
}

int TSerial::Rx(unsigned char* queue, int queue_size, timeval* timeout){
#ifdef WIN32
  COMMTIMEOUTS timo;
  DWORD count;

  timo.ReadIntervalTimeout = 0;
  timo.ReadTotalTimeoutMultiplier = 0;
  timo.ReadTotalTimeoutConstant = timeout->tv_sec * 1000 +
    timeout->tv_usec / 1000;
  timo.WriteTotalTimeoutMultiplier = 1;
  timo.WriteTotalTimeoutConstant = 1000;
  SetCommTimeouts (serline, &timo);

  if (! ReadFile (serline, queue, queue_size, &count, 0))
    throw Error_Device("Programmer is not responding.");
  return count;
#else
  int ret;
  fd_set rfds;
  FD_ZERO(&rfds); FD_SET(serline,&rfds);
  if ((ret=select(serline+1,&rfds,NULL,NULL, timeout))==-1){
    throw Error_C();
  }
  if (ret==0){throw Error_Device("Programmer is not responding.");}
  int size = read(serline, queue, queue_size);  
  return size;
#endif
}

int TSerial::Send(unsigned char* queue, int queue_size, int rec_queue_size){
  Tx(queue, queue_size);
  struct timeval time_out;
  time_out.tv_sec = 0;
  time_out.tv_usec = 100000;  
  if (rec_queue_size==-1){rec_queue_size = queue_size;}
  int total_len=0;  
  while(total_len<rec_queue_size){
    total_len += Rx(&queue[total_len], rec_queue_size - total_len, &time_out);
  }
  return total_len;
}

/* Constructor/Destructor
*/

TSerial::TSerial(){
#ifdef WIN32
#define B1200	CBR_1200
#define B2400	CBR_2400
#define B4800	CBR_4800
#define B9600	CBR_9600
#define B19200	CBR_19200
#define B38400	CBR_38400
#define B57600	CBR_57600
  DCB dcb;
  COMMTIMEOUTS timo;
  const char* dev_name = "COM2";
#else
  struct termios pmode;
  const char* dev_name = "/dev/avr";
#endif
  const char* val;
  speed_t speed = B19200;	/* default speed */
  
  struct TSpeed{
    const char* arg;
    speed_t speed;
  };
  const TSpeed speed_array[] = {
    {"1200", B1200},
    {"2400", B2400},
    {"4800", B4800},
    {"9600", B9600},
    {"19200", B19200},
    {"38400", B38400},
    {"57600", B57600},
    {"", 0}
  };
  
  /* Parse Command Line Parameters */
  if ((val=GetCmdParam("-dserial"))){dev_name = val;}
  if ((val=GetCmdParam("-dspeed"))){
    const TSpeed* speed_item = speed_array;
    for (;speed_item->arg[0] != 0; speed_item++){
      if (!strcmp(speed_item->arg, val)){speed = speed_item->speed;}
      break;
    } 
    if (speed_item->arg[0]==0){throw Error_Device("-dspeed: Invalid speed.");}
  }
  
  /* Open port and set serial attributes */
#ifdef WIN32
  serline = CreateFile (dev_name, GENERIC_READ | GENERIC_WRITE,
    0, 0, OPEN_EXISTING, 0, 0);
  if (serline == INVALID_HANDLE_VALUE)
    throw Error_C();

  memset (&dcb, 0, sizeof(dcb));
  GetCommState (serline, &dcb);
  dcb.BaudRate = speed;
  dcb.ByteSize = 8;
  dcb.StopBits = ONESTOPBIT;
  dcb.Parity = 0;
  dcb.fParity = FALSE;
  dcb.fOutX = FALSE;
  dcb.fInX = FALSE;
  dcb.fOutxCtsFlow = FALSE;
  dcb.fOutxDsrFlow = FALSE;
  dcb.fRtsControl = RTS_CONTROL_ENABLE;
  dcb.fNull = FALSE;
  dcb.fAbortOnError = FALSE;
  dcb.fBinary = TRUE;
  SetCommState (serline, &dcb);

  timo.ReadIntervalTimeout = 0;
  timo.ReadTotalTimeoutMultiplier = 0;
  timo.ReadTotalTimeoutConstant = 200;
  timo.WriteTotalTimeoutMultiplier = 1;
  timo.WriteTotalTimeoutConstant = 1000;
  SetCommTimeouts (serline, &timo);
#else
  if ((serline = open(dev_name, O_RDWR | O_NDELAY)) < 0 ){throw Error_C();}  
  tcgetattr(serline, &pmode);
  pmode.c_iflag = IGNPAR;	/* IGN_BRK not needed */
  pmode.c_oflag = 0;
  pmode.c_cflag = CS8 | CLOCAL | CREAD;   
  pmode.c_lflag = 0;
  pmode.c_cc [VMIN] = 1;
  pmode.c_cc [VTIME] = 0; 
  cfmakeraw(&pmode);
  cfsetispeed(&pmode, speed);
  cfsetospeed(&pmode, speed);
  tcsetattr(serline, TCSANOW, &pmode);
  
  /* Reopen port */
  int fd = serline;
  if ((serline = open(dev_name, O_RDWR)) < 0){throw Error_C();}
  close(fd);
#endif
}

TSerial::~TSerial(){
#ifdef WIN32
  CloseHandle(serline);
#else
  close(serline);
#endif
}
